#! /usr/bin/env python

## Copyright 1999-2001 by LivingLogic AG, Bayreuth, Germany.
## Copyright 1999-2001 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVINGLOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""URL class.

This module contains only one useful variable: the URL class
"""

__version__ = tuple(map(int, "$Revision: 2.19 $"[11:-2].split(".")))
# $Source: /data/cvsroot/LivingLogic/xist/_xist/url.py,v $

try:
	import Image
except ImportError:
	Image = None

import os, stat, types, urlparse, urllib, UserString

from xist import options

urlparse.uses_relative.append("root")
urlparse.uses_relative.append(None)
urlparse.uses_params.append("root")
urlparse.uses_query.append("root")
urlparse.uses_fragment.append("root")
urlparse.non_hierarchical.append("javascript")

def _normalize(path):
	"""
	normalize the path by removing combinations of down/up,
	empty path components or '.'.
	"""

	if path=="" or path==".":
		return ""

	wasAbsolutePath = (path[:1] == u"/") # the first element of absolute paths '', preserve it
	path = path.split("/")

	newpath = []
	for dir in path[:-1]:
		if dir and dir != ".":
			if dir == u".." and newpath and newpath[-1] != "..":
				newpath.pop()
			else:
				newpath.append(dir)
	newpath.append(path[-1])
	if wasAbsolutePath:
		newpath.insert(0, "")
	newpath = "/".join(newpath)

	return newpath

class URL(UserString.UserString):
	"""
	This class represents XSC URLs.
	Every instance has the following instance variables:
	scheme -- The scheme (e.g. "http" or "ftp"); there's a special scheme "root" for URLs relative to the current directory
	server -- The server name
	port -- The port number
	path -- The path to the file as a list of strings
	file -- The filename including the extension
	params -- The parametes
	query -- The query
	fragment -- The fragment
	These variables form a URL in the following way
	<scheme>://<server>:<port>/<path>/<file>;<params>?<query>#<fragment>

	There is one additional feature supported by this class: the scheme root
	is used for URLs that are relative to the current directory.
	"""

	__safe = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_,.-:/"

	def __init__(self, url):
		if isinstance(url, URL):
			url = url.data
		if url.startswith("file:"):
			url = url[5:]
			url = os.path.expanduser(os.path.normpath(url))
			url = "file:" + url
		(scheme, server, path, params, query, fragment) = urlparse.urlparse(url)
		self.scheme = scheme.lower() or None
		self.server = server.lower() or None
		self.path = _normalize(path)
		self.params = params or None
		self.query = query or None
		self.fragment = fragment or None
		url = urlparse.urlunparse((scheme, server, path, params, query, fragment))
		UserString.UserString.__init__(self, url)

	def __repr__(self):
		return "URL(%r)" % self.data

	def asString(self):
		return self._asString(0)

	def asPlainString(self):
		return self._asString(1)

	def _asString(self, plain):
		scheme = self.scheme
		if plain and self.scheme == u"root":
			scheme = u""
		return urlparse.urlunparse((scheme, self.server, self.path, self.params or u"", self.query or u"", self.fragment or u""))

	def __div__(self, other):
		"""Join a base URL and a possibly relative URL to form
		an absolute interpretation of the latter."""

		other = URL(other) # this implies clone/coerce

		if other.scheme==u"root":
			return other
		return URL(urlparse.urljoin(self.asString(), other.asString()))

	def clone(self):
		"""
		returns an identical clone of this URL.
		"""
		return URL(self)

	def isRemote(self):
		if self.scheme is None or self.scheme=="root":
			return 0
		else:
			return 1

	def relativeTo(self, other):
		"""
		<par nointent>returns <self/> interpreted relative
		to <code>other</code>, i.e. return a mimimal URL <code>rel</code>,
		such that <code>other/rel == <self/></code>.</par>
		"""
		new = self.clone() # make a copy and modify it
		if new.scheme in urlparse.uses_relative and (other.scheme==new.scheme or (new.scheme=="root" and other.scheme is None)) and other.server==new.server: # test if the URL uses the same scheme and goes to the same machine
			# if yes, the url is a relative one
			new.scheme = None
			new.server = None
			# we construct the path from one file to the other
			otherpath = other.path.split("/")
			otherfile = otherpath[-1]
			del otherpath[-1]
			newpath = new.path.split("/")
			newfile = newpath[-1]
			del newpath[-1]
			while len(otherpath) and len(newpath) and otherpath[0]==newpath[0]:
				# throw away identical directories in both paths (we don't have to go up
				# one dir in one path and down in the other for these identical directories)
				del otherpath[0]
				del newpath[0]
			# now for the rest of the path we have to go
			# up from file and down to path (the
			# directories for this are still in path)
			newpath[:0] = [u".."]*len(otherpath)
			if not len(newpath):
				# the link goes to the same URL ... we can throw away the filename, if we have a fragment in it
				if new.fragment is not None and newfile==otherfile and new.query==other.query and new.params==other.params:
					newfile = ""
			newpath.append(newfile)
			new.path = "/".join(newpath)
		return new

	def __cmp__(self, other):
		assert isinstance(other, URL), "URLs can only be compared with URLs"
		return cmp(self.scheme, other.scheme) or cmp(self.server, other.server) or cmp(self.path, other.path) or cmp(self.params, other.params) or cmp(self.query, other.query) or cmp(self.fragment, other.fragment)

	def __nonzero__(self):
		"""
		return if the URL is not empty (i.e. URL())
		"""
		return self.scheme is not None or self.server is not None or self.path != "" or self.params is not None or self.query is not None or self.fragment is not None

	def open(self):
		return urllib.urlopen(self.__quote())

	def retrieve(self):
		return urllib.urlretrieve(self.__quote())

	def info(self):
		file = urllib.urlopen(self.__quote())
		info = file.info()
		file.close()
		return info

	def read(self):
		file = self.open()
		data = file.read()
		file.close()
		return data

	def readlines(self):
		file = self.open()
		data = file.readlines()
		file.close()
		return data

	def __quote(self):
		"""
		encodes the URL with % escapes
		"""
		url = self.asPlainString().encode("utf8")
		#v = []
		#for c in url:
		#	if c in self.__safe:
		#		v.append(c)
		#	else:
		#		v.append("%%%02x" % ord(c))
		#return "".join(v)
		return url

	def isRetrieve(self):
		"""
		returns if the file specified by the URL should be retrieved or not
		according to the "remoteness" of the URL and the retrieve options.
		"""
		remote = self.isRemote()
		if (options.retrieveremote and remote) or (options.retrievelocal and (not remote)):
			return 1
		else:
			return 0

	def fileSize(self):
		"""
		returns the size of a file in bytes or None if the file shouldn't be read
		"""
		size = None
		if self.isRetrieve():
			try:
				(filename, headers) = self.retrieve()
				size = os.stat(filename)[stat.ST_SIZE]
			finally:
				urllib.urlcleanup()
		return size

	def imageSize(self):
		"""
		returns the size of an image as a tuple or None if the image shouldn't be read
		"""
		size = None
		if Image is not None:
			if self.isRetrieve():
				try:
					(filename, headers) = self.retrieve()
					if headers.maintype == "image":
						img = Image.open(filename)
						size = img.size
						del img
				finally:
					urllib.urlcleanup()
		return size

def test_normalize(input, output):
	"""
	Tests whether '_normalize' returns the expected results.
	"""
	test = _normalize(input)
	if test != output:
		raise str('normalize test failed. norm(%r)= %r != %r' % (input, test, output))

def test_url(o_lhs, o_rhs):
	"""
	Test whether instantiation, '/' and others have the expected effect.

	Check is done via comparing against another URL or a string; in the
	later case, the URL is converted using asString().
	"""
	import types
	lhs = o_lhs
	rhs = o_rhs
	if type(lhs) != type(rhs):
		if type(rhs) is types.StringType:
			if isinstance(lhs, URL):
				lhs = o_lhs.asString()
			else:
				lhs = str(lhs)
		elif type(lhs) is types.StringType:
			if isinstance(rhs, URL):
				rhs = o_rhs.asString()
			else:
				rhs = str(rhs)
	if lhs != rhs:
		print str('Test failed. %r (%r) != %r (%r)' % (o_lhs, lhs, o_rhs, rhs))

def test_relativeTo(from_, to, should):
	"""
	Test correctness of relativeTo().
	"""
	from_ = URL(from_)
	to = URL(to)
	
	relu = to.relativeTo(from_)
	u = relu.asString()
	if u != should:
		print str('Test failed: %r.relativeTo(%r) is %r but should be %r' % (to, from_, u, should))
	if from_/relu != to:
		print str('Invariance test failed: %r/%r=%r != %r' % (from_, relu, from_/relu, to))

test_input = urlparse.test_input

def test_url2():
	base = __empty = URL('')
	for line in test_input.splitlines():
		words = line.split()
		if not words:
			continue
		url = words[0]
		parts = URL(url)
		print '%-10s : %s' % (url, parts)
		abs = base/url
		if base is __empty:
			base = abs
		wrapped = '<URL:%s>' % abs.asString()
		print '%-10s = %s' % (url, wrapped)
		if len(words) == 3 and words[1] == '=':
			if wrapped != words[2]:
				print 'EXPECTED', words[2], '!!!!!!!!!!'

if __name__ == '__main__':
	test_normalize('', '')
	test_normalize('./', '')
	test_normalize('/./', '/')
	test_normalize('xx', 'xx')
	test_normalize('xx/yy', 'xx/yy')
	test_normalize('xx/../', '')
	test_normalize('xx/.././', '')
	test_normalize('./xx/../', '')
	test_normalize('./xx/.././', '')
	test_normalize('xx/./../', '')
	test_normalize('xx/yy/../', 'xx/')
	test_normalize('xx//yy/../../', '')
	test_normalize('xx//yy/./../', 'xx/')
	test_normalize('xx//yy//../', 'xx/')
	test_normalize('xx/../..//', '../')
	test_normalize('xx/.././../', '../')
	test_normalize('xx/.', 'xx/.')
	test_normalize('xx/./', 'xx/')
	test_normalize('./xx', 'xx')
	test_normalize('/xx', '/xx')
	test_normalize('/./xx', '/xx')
	print 'test passed: _normalize'

	test_url(URL("."), ".") # "." by itself will not be treated as a directory name
	test_url(URL("./"), "./")
	test_url(URL(".."), "..") # ".." by itself will not be treated as a directory name
	test_url(URL("../"), "../")
	test_url(URL("http://aa/bb/cc.html"), "http://aa/bb/cc.html")
	test_url(URL("/aa/bb/cc.html"), "/aa/bb/cc.html")
	test_url(URL("http:bb/cc/"), "http:bb/cc/")
	test_url(URL("http:/bb/cc/"), "http:/bb/cc/")
	test_url(URL("http:") / URL("/bb/cc/"), "http:/bb/cc/")
	test_url(URL("http://test.com/index.html") / URL("impress.html"), "http://test.com/impress.html")
	test_url(URL("/bb/cc/") / URL("http:"), "http:")
	test_url(URL("mailto:x@y.z") / URL("index.html"), "index.html")
	test_url(URL("javascript: return ':/:/:';") / URL("index.html"), "index.html")
	test_url(URL("http://test.com/gurk/hurz.gif") / URL("/index.html"), "http://test.com/index.html")
	test_url(URL("http://test.com/gurk/hurz.gif") / URL("../"), "http://test.com/")
	test_url(URL("http://test.com/gurk/hurz.gif") / URL("../gurk.gif?foo=bar#nix"), "http://test.com/gurk.gif?foo=bar#nix")
	test_url(URL("http://test.com/gurk/hurz.gif") / URL("../../gurk.gif?foo=bar#nix"), "http://test.com/../gurk.gif?foo=bar#nix")
	test_url(URL("http://test.com/gurk/hurz.gif") / URL("root:gurk.gif"), "root:gurk.gif")
	test_url(URL("root:gurk.gif") / URL("http://test.com/gurk/hurz.gif"), "http://test.com/gurk/hurz.gif")
	test_url(URL("root:gurk/hurz/hinz.gif") / URL("hinz/kunz.gif"), "root:gurk/hurz/hinz/kunz.gif")
	test_url(URL("root:gurk/hurz/hinz.gif") / URL("root:hinz/kunz.gif"), "root:hinz/kunz.gif")
	test_url(URL("#mark"), "#mark")
	print 'test passed: URL()'

	test_relativeTo('', '', '')
	test_relativeTo('cc.html', '', '')
	test_relativeTo('./cc.html', '', '')
	test_relativeTo("root:xist/Documentation.html", "http://www.livinglogic.de/", "http://www.livinglogic.de/")
	test_relativeTo('root:cc.html', 'root:', '')
	test_relativeTo('root:cc.html', '', '')
	test_relativeTo('cc.html', '#mark', '#mark')
	test_relativeTo('root:cc.html', 'root:#mark', './#mark')
	test_relativeTo('root:cc.html', '#mark', '#mark')
	test_relativeTo('root:cc.html', 'root:cc.html#mark', '#mark')
	test_relativeTo('root:cc.html', 'root:dd.html#mark', 'dd.html#mark')
	test_relativeTo('root:aa/bb/cc.html', 'root:', '../../')
	#test_relativeTo('', '', '')
	test_relativeTo('http://server/aa/bb.html', 'http://server/aa/cc.html', 'cc.html')
	test_relativeTo('/aa/bb.html', '/xx.html', '../xx.html')
	print 'test passed: relativeTo()'

	test_url2()
